// WinHotkeyCtrl.cpp

#include "stdafx.h"
#include "vkCodes.h"
#include "WinHotkeyCtrl.h"

using namespace NVkCodes;

//-----------------------------------------------------------------

namespace NWinHotkeyCtrl {

//-----------------------------------------------------------------

#define WM_KEY                      (WM_USER + 444)

#define MAKEWHCDATA(vkCode, fModSet, fModRel, fIsPressed) \
    ((DWORD)(((BYTE)((DWORD_PTR)(vkCode) & 0xff)) | \
    (((DWORD)((BYTE)((DWORD_PTR)(fModSet) & 0xff))) << 8) | \
    (((DWORD)((BYTE)((DWORD_PTR)(fModRel) & 0xff))) << 16) | \
    (((DWORD)((BYTE)((DWORD_PTR)(fIsPressed) & 0xff))) << 24)))

/*****************************************************************/
/*                                                               */
/*  DWORD dwWhcData = GetWindowLongPtr(hwndWhc, GWLP_USERDATA);  */
/*                                                               */
/*  LOBYTE(LOWORD(dwWhcData)) - key virtual code                 */
/*  HIBYTE(LOWORD(dwWhcData)) - modifiers set                    */
/*  LOBYTE(HIWORD(dwWhcData)) - modifiers released               */
/*  HIBYTE(HIWORD(dwWhcData)) - key is pressed(?)                */
/*                                                               */
/*****************************************************************/


LRESULT CALLBACK _WinHotkeyCtrlProc(HWND, UINT, WPARAM, LPARAM);
void _SetWhcText(HWND, DWORD);

//-----------------------------------------------------------------

WNDPROC _wpEditProc                 = NULL;
HHOOK _hhookKb                      = NULL;
HWND _hwndWhc                       = NULL;

//-----------------------------------------------------------------

BOOL InitWinHotkeyCtrls() {
    
    WNDCLASSEX wcex;
    wcex.cbSize = sizeof(WNDCLASSEX);
    if (!GetClassInfoEx(GetModuleHandle(NULL), _T("Edit"), &wcex))
        return(FALSE);
    _wpEditProc = wcex.lpfnWndProc;

    return(_wpEditProc != NULL);
}

//-----------------------------------------------------------------

BOOL SubClassWinHotkeyCtrl(HWND hwndWhc) {

    _ASSERT(hwndWhc);

    //_ASSERT(_wpEditProc);
    if (!_wpEditProc)
        if (!InitWinHotkeyCtrls())
            return(FALSE);
    
    SetWindowLongPtr(hwndWhc, GWLP_WNDPROC, (LONG)(LONG_PTR)(WNDPROC)_WinHotkeyCtrlProc);
    SetWinHotkey(hwndWhc, MAKEWHCDATA(0, 0, 0, 0));

    return(TRUE);
}

//-----------------------------------------------------------------

#if _WIN32_WINNT < 0x500

LRESULT CALLBACK _KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam) {

    if (nCode == HC_ACTION) {
        PostMessage(_hwndWhc, WM_KEY, wParam, (lParam & 0x80000000));
    }
    return(1);
}

#else // _WIN32_WINNT >= 0x500

LRESULT CALLBACK _LowLevelKeyboardProc(int nCode, WPARAM wParam, LPARAM lParam) {

    if (nCode == HC_ACTION && (wParam == WM_KEYDOWN || 
        wParam == WM_SYSKEYDOWN || wParam == WM_KEYUP || wParam == WM_SYSKEYUP)) {

        PostMessage(_hwndWhc, WM_KEY, 
            ((PKBDLLHOOKSTRUCT)lParam)->vkCode, (wParam & 1));          
    }
    return(1);
}

#endif // _WIN32_WINNT >= 0x500

//-----------------------------------------------------------------

BOOL _UninstallKbHook() {

    BOOL fOk = FALSE;
    if (_hhookKb) {
        fOk = UnhookWindowsHookEx(_hhookKb);
        _hhookKb = NULL;
    }
    _hwndWhc = NULL;
    return(fOk);
}

//-----------------------------------------------------------------

BOOL _InstallKbHook(HWND hwndWhc) {

    if (_hhookKb)
        _UninstallKbHook();

    _hwndWhc = hwndWhc;

#if _WIN32_WINNT < 0x500
    _hhookKb = SetWindowsHookEx(WH_KEYBOARD, 
        (HOOKPROC)_KeyboardProc, NULL, GetCurrentThreadId());
#else // _WIN32_WINNT >= 0x500
    _hhookKb = SetWindowsHookEx(WH_KEYBOARD_LL, 
        (HOOKPROC)_LowLevelKeyboardProc, GetModuleHandle(NULL), NULL);
#endif // _WIN32_WINNT >= 0x500

    return(_hhookKb != NULL);
}

//-----------------------------------------------------------------

void _SetWhcText(HWND hwndWhc, DWORD dwHk) {

    TCHAR pszText[64];
    if (HotkeyToString(dwHk, pszText, sizeof(pszText) / sizeof(pszText[0])))
        SetWindowText(hwndWhc, pszText);
    else SetWindowText(hwndWhc, _T("None"));
    SendMessage(hwndWhc, EM_SETSEL, 0x8fffffff, 0x8fffffff);
}

//-----------------------------------------------------------------

DWORD GetWinHotkey(HWND hwndWhc) {

    _ASSERT(hwndWhc);
    return(LOWORD(GetWindowLongPtr(hwndWhc, GWLP_USERDATA)));
}

//-----------------------------------------------------------------

void SetWinHotkey(HWND hwndWhc, DWORD dwHk) {

    DWORD dwWhcData = MAKEWHCDATA(LOBYTE(LOWORD(dwHk)), 
        HIBYTE(LOWORD(dwHk)), HIBYTE(LOWORD(dwHk)), FALSE);
    SetWindowLongPtr(hwndWhc, GWLP_USERDATA, dwWhcData);
    _SetWhcText(hwndWhc, dwWhcData);
}

//-----------------------------------------------------------------

LRESULT CALLBACK _WinHotkeyCtrlProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {

    switch (uMsg) {

        case WM_KEY: {

            DWORD dwWhcData = (DWORD)GetWindowLongPtr(hwnd, GWLP_USERDATA);

            DWORD vkCode    = LOBYTE(LOWORD(dwWhcData));
            DWORD fModSet   = HIBYTE(LOWORD(dwWhcData));
            DWORD fModRel   = LOBYTE(HIWORD(dwWhcData));
            BOOL fIsPressed = HIBYTE(HIWORD(dwWhcData));

            DWORD fMod = 0;
            BOOL fRedraw = TRUE;

            switch (wParam) {
                case VK_LWIN:
                case VK_RWIN:       fMod = MOD_WIN; break;
                case VK_CONTROL:
                case VK_LCONTROL:
                case VK_RCONTROL:   fMod = MOD_CONTROL; break;
                case VK_MENU:
                case VK_LMENU: 
                case VK_RMENU:      fMod = MOD_ALT; break;
                case VK_SHIFT:
                case VK_LSHIFT:
                case VK_RSHIFT:     fMod = MOD_SHIFT; break;
            }

            if (fMod) { // modifier
                if (!lParam) { // press
                    if(!fIsPressed && vkCode) {
                        fModSet = fModRel = 0;
                        vkCode = 0;
                    } 
                    fModRel &= ~fMod;
                } else if (fModSet & fMod) // release
                    fModRel |= fMod;

                if (fIsPressed || !vkCode) {
                    if (!lParam) { // press
                        if (!(fModSet & fMod)) { // new modifier
                            fModSet |= fMod;
                        } else
                            fRedraw = FALSE;
                    } else fModSet &= ~fMod;
                }
            } else { // another key

                if (wParam == VK_DELETE && fModSet == (MOD_CONTROL | MOD_ALT)) {
                    fModSet = fModRel = 0; // skip "Ctrl + Alt + Del"
                    vkCode = 0;
                    fIsPressed = FALSE;
                } else if (wParam == vkCode && lParam) {
                    fIsPressed = FALSE;
                    fRedraw = FALSE;
                } else {
                    if (!fIsPressed && !lParam) { // pressed a another key
                        if (fModRel & fModSet) {
                            fModSet = fModRel = 0;
                        }
                        vkCode = (DWORD)wParam;
                        fIsPressed = TRUE;
                    }
                }
            }
            dwWhcData = MAKEWHCDATA(vkCode, fModSet, fModRel, fIsPressed);
            SetWindowLongPtr(hwnd, GWLP_USERDATA, (LONG)dwWhcData);

            if (fRedraw)
                _SetWhcText(hwnd, dwWhcData);

            return(0);
                     }

        case WM_SETFOCUS:
            _InstallKbHook(hwnd);
            break;

        case WM_KILLFOCUS:
            _UninstallKbHook();
            break;
            
        case WM_CONTEXTMENU: {

            UINT id;
            HMENU hmenu, hmenu2;
            hmenu = CreatePopupMenu();

#if _WIN32_WINNT >= 0x500
            hmenu2 = CreatePopupMenu();
            for (id = VK_BROWSER_BACK; id <= VK_LAUNCH_APP2; id++)
                AppendMenu(hmenu2, MF_STRING, id, GetKeyName(id));
            AppendMenu(hmenu, MF_STRING | MF_POPUP, (UINT_PTR)hmenu2, _T("Multimedia"));
#endif // _WIN32_WINNT >= 0x500

            hmenu2 = CreatePopupMenu();
            AppendMenu(hmenu2, MF_STRING, VK_RETURN, GetKeyName(VK_RETURN));
            AppendMenu(hmenu2, MF_STRING, VK_ESCAPE, GetKeyName(VK_ESCAPE));
            AppendMenu(hmenu2, MF_STRING, VK_TAB, GetKeyName(VK_TAB));
            AppendMenu(hmenu2, MF_STRING, VK_CAPITAL, GetKeyName(VK_CAPITAL));
            AppendMenu(hmenu2, MF_STRING, VK_BACK, GetKeyName(VK_BACK));
            AppendMenu(hmenu2, MF_STRING, VK_INSERT, GetKeyName(VK_INSERT));
            AppendMenu(hmenu2, MF_STRING, VK_DELETE, GetKeyName(VK_DELETE));
            for (id = VK_SPACE; id <= VK_DOWN; id++)
                AppendMenu(hmenu2, MF_STRING, id, GetKeyName(id));
            AppendMenu(hmenu, MF_STRING | MF_POPUP, (UINT_PTR)hmenu2, _T("Standard"));

            hmenu2 = CreatePopupMenu();
            for (id = VK_F1; id <= VK_F24; id++)
                AppendMenu(hmenu2, MF_STRING, id, GetKeyName(id));
            AppendMenu(hmenu, MF_STRING | MF_POPUP, (UINT_PTR)hmenu2, _T("Functionality"));


            DWORD dwWhcData = (DWORD)GetWindowLongPtr(hwnd, GWLP_USERDATA);
            DWORD vkCode = LOBYTE(LOWORD(dwWhcData));
            DWORD fModSet = HIBYTE(LOWORD(dwWhcData));
            DWORD fModRel = LOBYTE(HIWORD(dwWhcData));
            BOOL fIsPressed = HIBYTE(HIWORD(dwWhcData));

            AppendMenu(hmenu, MF_SEPARATOR, 0, NULL);
            AppendMenu(hmenu, (fModSet & MOD_WIN) ? 
                (MF_STRING | MF_CHECKED) : MF_STRING, VK_LWIN, _T("Win-key"));
            AppendMenu(hmenu, (fModSet & MOD_CONTROL) ? 
                (MF_STRING | MF_CHECKED) : MF_STRING, VK_CONTROL, _T("Control-key"));
            AppendMenu(hmenu, (fModSet & MOD_SHIFT) ? 
                (MF_STRING | MF_CHECKED) : MF_STRING, VK_SHIFT, _T("Shift-key"));
            AppendMenu(hmenu, (fModSet & MOD_ALT) ? 
                (MF_STRING | MF_CHECKED) : MF_STRING, VK_MENU, _T("Alt-key"));
            
            UINT uMenuID = TrackPopupMenu(hmenu, 
                TPM_RIGHTALIGN | TPM_RIGHTBUTTON | TPM_NONOTIFY | TPM_RETURNCMD,
                LOWORD(lParam), HIWORD(lParam), 0, hwnd, NULL);

            if (uMenuID && uMenuID < 256) {
                switch (uMenuID) {
                    case VK_LWIN:
                        if (vkCode) {
                            fModSet ^= MOD_WIN;
                            fModRel |= fModSet & MOD_WIN;
                        }
                        break;
                    case VK_CONTROL:
                        if (vkCode) {
                            fModSet ^= MOD_CONTROL;
                            fModRel |= fModSet & MOD_CONTROL;
                        }
                        break;
                    case VK_SHIFT:
                        if (vkCode) {
                            fModSet ^= MOD_SHIFT;
                            fModRel |= fModSet & MOD_SHIFT;
                        }
                        break;
                    case VK_MENU:
                        if (vkCode) {
                            fModSet ^= MOD_ALT;
                            fModRel |= fModSet & MOD_ALT;
                        }
                        break;

                    default:
                        vkCode = uMenuID;
                        fIsPressed = FALSE;
                        break;
                }
                dwWhcData = MAKEWHCDATA(vkCode, fModSet, fModRel, fIsPressed);
                SetWindowLongPtr(hwnd, GWLP_USERDATA, (LONG)dwWhcData);
                _SetWhcText(hwnd, dwWhcData);
                SetFocus(hwnd);
            }
            DestroyMenu(hmenu);
            return(0);
                             }

        case WM_CHAR:
        case WM_SETCURSOR:
            return(0);

        case WM_DESTROY:
            if (hwnd == _hwndWhc)
                _UninstallKbHook();
            break;
    }
    return(CallWindowProc(_wpEditProc, hwnd, uMsg, wParam, lParam));
}

//-----------------------------------------------------------------

} // namespace NWinHotkeyCtrl

//-----------------------------------------------------------------
